@@ -9,10 +9,14 @@ Baron's unofficial draft (https://dbaron.org/css/intrinsic/).
 """
 
 import sys
+from functools import cache
 from math import inf
 
+from ..css import resolve_math
+from ..css.functions import check_math
+from ..css.validation import validate_non_shorthand
 from ..formatting_structure import boxes
-from ..text.line_break import split_first_line
+from ..text.line_break import can_break_text, split_first_line
 from .replaced import default_image_sizing
 
 
@@ -42,21 +46,21 @@ def min_content_width(context, box, outer=True):
         return table_and_columns_preferred_widths(context, box, outer)[0]
     elif isinstance(box, boxes.TableCellBox):
         return table_cell_min_content_width(context, box, outer)
-    elif isinstance(box, (
-            boxes.BlockContainerBox, boxes.TableColumnBox, boxes.FlexBox)):
+    elif isinstance(box, (boxes.BlockContainerBox, boxes.TableColumnBox)):
         return block_min_content_width(context, box, outer)
     elif isinstance(box, boxes.TableColumnGroupBox):
         return column_group_content_width(context, box)
     elif isinstance(box, (boxes.InlineBox, boxes.LineBox)):
-        return inline_min_content_width(
-            context, box, outer, is_line_start=True)
+        return inline_min_content_width(context, box, outer, is_line_start=True)
     elif isinstance(box, boxes.ReplacedBox):
         return replaced_min_content_width(box, outer)
     elif isinstance(box, boxes.FlexContainerBox):
         return flex_min_content_width(context, box, outer)
+    elif isinstance(box, boxes.GridContainerBox):
+        # TODO: Get real grid size.
+        return block_min_content_width(context, box, outer)
     else:
-        raise TypeError(
-            f'min-content width for {type(box).__name__} not handled yet')
+        raise TypeError(f'min-content width for {type(box).__name__} not handled yet')
 
 
 def max_content_width(context, box, outer=True):
@@ -68,28 +72,28 @@ def max_content_width(context, box, outer=True):
     if box.is_table_wrapper:
         return table_and_columns_preferred_widths(context, box, outer)[1]
     elif isinstance(box, boxes.TableCellBox):
-        return table_cell_max_content_width(context, box, outer)
-    elif isinstance(box, (
-            boxes.BlockContainerBox, boxes.TableColumnBox, boxes.FlexBox)):
+        return table_cell_min_max_content_width(context, box, outer)[1]
+    elif isinstance(box, (boxes.BlockContainerBox, boxes.TableColumnBox)):
         return block_max_content_width(context, box, outer)
     elif isinstance(box, boxes.TableColumnGroupBox):
         return column_group_content_width(context, box)
     elif isinstance(box, (boxes.InlineBox, boxes.LineBox)):
-        return inline_max_content_width(
-            context, box, outer, is_line_start=True)
+        return inline_max_content_width(context, box, outer, is_line_start=True)
     elif isinstance(box, boxes.ReplacedBox):
         return replaced_max_content_width(box, outer)
     elif isinstance(box, boxes.FlexContainerBox):
         return flex_max_content_width(context, box, outer)
+    elif isinstance(box, boxes.GridContainerBox):
+        # TODO: Get real grid size.
+        return block_max_content_width(context, box, outer)
     else:
-        raise TypeError(
-            f'max-content width for {type(box).__name__} not handled yet')
+        raise TypeError(f'max-content width for {type(box).__name__} not handled yet')
 
 
 def _block_content_width(context, box, function, outer):
     """Helper to create ``block_*_content_width.``"""
     width = box.style['width']
-    if width == 'auto' or width.unit == '%':
+    if width == 'auto' or check_math(width) or width.unit == '%':
         # "percentages on the following properties are treated instead as
         # though they were the following: width: auto"
         # https://dbaron.org/css/intrinsic/#outer-intrinsic
@@ -98,7 +102,7 @@ def _block_content_width(context, box, function, outer):
             if not child.is_absolutely_positioned()]
         width = max(children_widths) if children_widths else 0
     else:
-        assert width.unit == 'px'
+        assert width.unit.lower() == 'px'
         width = width.value
 
     return adjust(box, outer, width)
@@ -108,11 +112,13 @@ def min_max(box, width):
     """Get box width from given width and box min- and max-widths."""
     min_width = box.style['min_width']
     max_width = box.style['max_width']
-    if min_width == 'auto' or min_width.unit == '%':
+    min_pending = check_math(min_width)
+    max_pending = check_math(max_width)
+    if min_width == 'auto' or min_pending or min_width.unit == '%':
         min_width = 0
     else:
         min_width = min_width.value
-    if max_width == 'auto' or max_width.unit == '%':
+    if max_width == 'auto' or max_pending or max_width.unit == '%':
         max_width = inf
     else:
         max_width = max_width.value
@@ -122,10 +128,12 @@ def min_max(box, width):
             1, box.style['font_size'])
         if ratio is not None:
             min_height = box.style['min_height']
-            if min_height != 'auto' and min_height.unit != '%':
-                min_width = max(min_width, min_height.value * ratio)
             max_height = box.style['max_height']
-            if max_height != 'auto' and max_height.unit != '%':
+            min_pending = check_math(min_height)
+            max_pending = check_math(max_height)
+            if min_height != 'auto' and not min_pending and min_height.unit != '%':
+                min_width = max(min_width, min_height.value * ratio)
+            if max_height != 'auto' and not min_pending and max_height.unit != '%':
                 max_width = min(max_width, max_height.value * ratio)
 
     return max(min_width, min(width, max_width))
@@ -145,7 +153,7 @@ def margin_width(box, width, left=True, right=True):
     ):
         style_value = box.style[value]
         if style_value != 'auto':
-            if style_value.unit == 'px':
+            if style_value.unit.lower() == 'px':
                 width += style_value.value
             else:
                 assert style_value.unit == '%'
@@ -235,7 +243,7 @@ def column_group_content_width(context, box):
     if width == 'auto' or width.unit == '%':
         width = 0
     else:
-        assert width.unit == 'px'
+        assert width.unit.lower() == 'px'
         width = width.value
 
     return adjust(box, False, width)
@@ -244,71 +252,76 @@ def column_group_content_width(context, box):
 def table_cell_min_content_width(context, box, outer):
     """Return the min-content width for a ``TableCellBox``."""
     # See https://www.w3.org/TR/css-tables-3/#outer-min-content
-    min_width = box.style['min_width']
-    if min_width == 'auto':
-        min_width = 0
-    else:
-        min_width = min_width.value
+    # The outer min-content width of a table-cell is
+    # max(min-width, min-content width) adjusted by
+    # the cell intrinsic offsets.
     children_widths = [
-        max(min_width, min_content_width(context, child))
+        min_content_width(context, child)
         for child in box.children
         if not child.is_absolutely_positioned()]
-    children_min_width = margin_width(
-        box, max(children_widths) if children_widths else 0)
+    children_min_width = adjust(
+        box,
+        outer,
+        max(children_widths) if children_widths else 0)
 
-    width = box.style['width']
-    if width != 'auto' and width.unit == 'px':
-        cell_min_width = adjust(box, outer, width.value)
-    else:
-        cell_min_width = 0
-
-    return max(children_min_width, cell_min_width)
+    return children_min_width
 
 
-def table_cell_max_content_width(context, box, outer):
-    """Return the max-content width for a ``TableCellBox``."""
-    return max(
-        table_cell_min_content_width(context, box, outer),
-        block_max_content_width(context, box, outer))
+def table_cell_min_max_content_width(context, box, outer=True):
+    """Return the min- and max-content width for a ``TableCellBox``."""
+    # This is much faster than calling min and max separately.
+    min_width = table_cell_min_content_width(context, box, outer)
+    max_width = max(min_width, block_max_content_width(context, box, outer))
+    return min_width, max_width
 
 
-def inline_line_widths(context, box, outer, is_line_start, minimum,
-                       skip_stack=None, first_line=False):
+def inline_line_widths(context, box, outer, is_line_start, minimum, skip_stack=None,
+                       first_line=False):
+    """Yield line width for each line."""
+
+    # Set text indent.
+    text_indent = 0
     if isinstance(box, boxes.LineBox):
-        if box.style['text_indent'].unit == '%':
-            # TODO: this is wrong, text-indent percentages should be resolved
-            # before calling this function.
-            text_indent = 0
-        else:
+        indent_token = box.style['text_indent']
+        if check_math(indent_token):
+            # Ignore percentages by setting refer_to to 0.
+            result = resolve_math(indent_token, box.style, 'text_indent', refer_to=0)
+            value = validate_non_shorthand((result,), 'text-indent')[0][1]
+            if value and value.unit != '%':
+                text_indent = value.value
+        elif indent_token.unit != '%':
             text_indent = box.style['text_indent'].value
-    else:
-        text_indent = 0
 
+    # Yield widths for each line.
     current_line = 0
     if skip_stack is None:
         skip = 0
     else:
         (skip, skip_stack), = skip_stack.items()
     for child in box.children[skip:]:
+        # Skip absolutely positioned elements.
         if child.is_absolutely_positioned():
-            continue  # Skip
+            continue
 
+        # None is used in "lines" to track line breaks, transformed to 0 when yielded.
         if isinstance(child, boxes.InlineBox):
+            # Inline box, call function recursively.
             lines = inline_line_widths(
-                context, child, outer, is_line_start, minimum, skip_stack,
-                first_line)
+                context, child, outer, is_line_start, minimum, skip_stack, first_line)
             if first_line:
-                lines = [next(lines)]
+                lines = [next(lines) or None]
             else:
-                lines = list(lines)
+                lines = [line or None for line in lines]
             if len(lines) == 1:
-                lines[0] = adjust(child, outer, lines[0])
+                lines[0] = adjust(child, outer, lines[0] or 0)
             else:
-                lines[0] = adjust(child, outer, lines[0], right=False)
-                lines[-1] = adjust(child, outer, lines[-1], left=False)
+                lines[0] = adjust(child, outer, lines[0] or 0, right=False) or None
+                lines[-1] = adjust(child, outer, lines[-1] or 0, left=False) or None
         elif isinstance(child, boxes.TextBox):
-            space_collapse = child.style['white_space'] in (
-                'normal', 'nowrap', 'pre-line')
+            # Text box, split into lines.
+            white_space = child.style['white_space']
+            space_collapse = white_space in ('normal', 'nowrap', 'pre-line')
+            text_wrap = white_space in ('normal', 'pre-wrap', 'pre-line')
             if skip_stack is None:
                 skip = 0
             else:
@@ -317,26 +330,30 @@ def inline_line_widths(context, box, outer, is_line_start, minimum,
             child_text = child.text.encode()[(skip or 0):]
             if is_line_start and space_collapse:
                 child_text = child_text.lstrip(b' ')
-            if minimum and child_text == b' ':
-                lines = [0, 0]
-            else:
-                max_width = 0 if minimum else None
-                lines = []
-                resume_index = new_resume_index = 0
-                while new_resume_index is not None:
-                    resume_index += new_resume_index
-                    _, _, new_resume_index, width, _, _ = (
-                        split_first_line(
-                            child_text[resume_index:].decode(), child.style,
-                            context, max_width, child.justification_spacing,
-                            is_line_start=is_line_start, minimum=True))
-                    lines.append(width)
-                    if first_line:
-                        break
-                if first_line and new_resume_index:
-                    current_line += lines[0]
+            max_width = 0 if minimum else None
+            lines = []
+            resume_index = new_resume_index = 0
+            while new_resume_index is not None:
+                resume_index += new_resume_index
+                _, _, new_resume_index, width, _, _ = split_first_line(
+                    child_text[resume_index:].decode(), child.style, context, max_width,
+                    child.justification_spacing, is_line_start=is_line_start,
+                    minimum=True)
+                lines.append(width or None)
+                if first_line:
                     break
+            if first_line and new_resume_index:
+                # We only need the first line, break early.
+                current_line += lines[0] or 0
+                break
+            # TODO: use the real next character instead of 'a' to detect line breaks.
+            last_letter = child_text.decode()[-1:]
+            can_break = can_break_text(last_letter + 'a', child.style['lang'])
+            if minimum and text_wrap and can_break:
+                # Add all possible line breaks for minimal width.
+                lines.append(None)
         else:
+            # Replaced elements, inline blocks…
             # https://www.w3.org/TR/css-text-3/#overflow-wrap
             # "The line breaking behavior of a replaced element
             #  or other atomic inline is equivalent to that
@@ -345,20 +362,20 @@ def inline_line_widths(context, box, outer, is_line_start, minimum,
             # "By default, there is a break opportunity
             #  both before and after any inline object."
             if minimum:
-                lines = [0, max_content_width(context, child), 0]
+                lines = [None, min_content_width(context, child), None]
             else:
                 lines = [max_content_width(context, child)]
-        # The first text line goes on the current line
-        current_line += lines[0]
+        # The first text line goes on the current line.
+        current_line += lines[0] or 0
         if len(lines) > 1:
-            # Forced line break
+            # Forced line break(s).
             yield current_line + text_indent
             text_indent = 0
             if len(lines) > 2:
                 for line in lines[1:-1]:
-                    yield line
-            current_line = lines[-1]
-        is_line_start = lines[-1] == 0
+                    yield line or 0
+            current_line = lines[-1] or 0
+        is_line_start = lines[-1] is None
         skip_stack = None
     yield current_line + text_indent
 
@@ -369,15 +386,18 @@ def _percentage_contribution(box):
     https://dbaron.org/css/intrinsic/#pct-contrib
 
     """
+    min_width = box.style['min_width']
     min_width = (
-        box.style['min_width'].value if box.style['min_width'] != 'auto' and
-        box.style['min_width'].unit == '%' else 0)
+        min_width.value if min_width != 'auto' and
+        not check_math(min_width) and min_width.unit == '%' else 0)
+    max_width = box.style['max_width']
     max_width = (
-        box.style['max_width'].value if box.style['max_width'] != 'auto' and
-        box.style['max_width'].unit == '%' else inf)
+        max_width.value if max_width != 'auto' and
+        not check_math(max_width) and max_width.unit == '%' else inf)
+    width = box.style['width']
     width = (
-        box.style['width'].value if box.style['width'] != 'auto' and
-        box.style['width'].unit == '%' else 0)
+        width.value if width != 'auto' and
+        not check_math(width) and width.unit == '%' else 0)
     return max(min_width, min(width, max_width))
 
 
@@ -437,8 +457,8 @@ def table_and_columns_preferred_widths(context, box, outer=True):
             box, outer=True, width=block_max_content_width(context, table))
         result = ([], [], [], [], total_horizontal_border_spacing, [])
         context.tables[table] = result = {
-            False: (min_width, max_width) + result,
-            True: (outer_min_width, outer_max_width) + result,
+            False: (min_width, max_width, *result),
+            True: (outer_min_width, outer_max_width, *result),
         }
         return result[outer]
 
@@ -457,39 +477,35 @@ def table_and_columns_preferred_widths(context, box, outer=True):
         break
 
     colspan_cells = []
+    colspans = set()
 
     # Define the intermediate content widths
-    min_content_widths = [0 for i in range(grid_width)]
-    max_content_widths = [0 for i in range(grid_width)]
-    intrinsic_percentages = [0 for i in range(grid_width)]
+    min_content_widths = [0] * grid_width
+    max_content_widths = [0] * grid_width
+    intrinsic_percentages = [0] * grid_width
 
     # Intermediate content widths for span 1
     for i in range(grid_width):
         for groups in (column_groups, columns):
-            if groups[i]:
+            if group := groups[i]:
                 min_content_widths[i] = max(
-                    min_content_widths[i],
-                    min_content_width(context, groups[i]))
+                    min_content_widths[i], min_content_width(context, group))
                 max_content_widths[i] = max(
-                    max_content_widths[i],
-                    max_content_width(context, groups[i]))
+                    max_content_widths[i], max_content_width(context, group))
                 intrinsic_percentages[i] = max(
-                    intrinsic_percentages[i],
-                    _percentage_contribution(groups[i]))
+                    intrinsic_percentages[i], _percentage_contribution(group))
         for cell in zipped_grid[i]:
-            if cell:
-                if cell.colspan == 1:
-                    min_content_widths[i] = max(
-                        min_content_widths[i],
-                        min_content_width(context, cell))
-                    max_content_widths[i] = max(
-                        max_content_widths[i],
-                        max_content_width(context, cell))
-                    intrinsic_percentages[i] = max(
-                        intrinsic_percentages[i],
-                        _percentage_contribution(cell))
-                else:
-                    colspan_cells.append(cell)
+            if not cell:
+                continue
+            if cell.colspan == 1:
+                min_width, max_width = table_cell_min_max_content_width(context, cell)
+                min_content_widths[i] = max(min_content_widths[i], min_width)
+                max_content_widths[i] = max(max_content_widths[i], max_width)
+                intrinsic_percentages[i] = max(
+                    intrinsic_percentages[i], _percentage_contribution(cell))
+            else:
+                colspan_cells.append(cell)
+                colspans.add(cell.colspan - 1)
 
     # Intermediate content widths for span > 1 is wrong in the 4.1 section, as
     # explained in its third issue. Min- and max-content widths are handled by
@@ -497,46 +513,49 @@ def table_and_columns_preferred_widths(context, box, outer=True):
     # widths to columns that have originating cells.
 
     # Intermediate intrinsic percentage widths for span > 1
-    for span in range(1, grid_width):
+    rows_origins = []
+    for y, row in enumerate(grid):
+        origin = None
+        rows_origins.append(row_origins := [])
+        for x, cell in enumerate(row):
+            if cell:
+                origin = x
+            row_origins.append(origin)
+
+    @cache
+    def get_percentage_contribution(origin_cell, origin, max_content_width):
+        # Cached for big colspan values, see #1155.
+        cell_slice = slice(origin, origin + origin_cell.colspan)
+        baseline_percentage = sum(intrinsic_percentages[cell_slice])
+        cell_percentage_contribution = _percentage_contribution(origin_cell)
+        diff = max(0, cell_percentage_contribution - baseline_percentage)
+        other_columns_contributions = [
+            max_content_widths[j]
+            for j in range(origin, origin + origin_cell.colspan)
+            if intrinsic_percentages[j] == 0]
+        other_columns_contributions_sum = sum(other_columns_contributions)
+        if other_columns_contributions_sum == 0:
+            ratio = 1 / (len(other_columns_contributions) or 1)
+        else:
+            ratio = max_content_width / other_columns_contributions_sum
+        return diff * ratio
+
+    for span in sorted(colspans):
         percentage_contributions = []
         for i in range(grid_width):
-            percentage_contribution = intrinsic_percentages[i]
-            for j, cell in enumerate(zipped_grid[i]):
-                indexes = [k for k in range(i + 1) if grid[j][k]]
-                if not indexes:
+            if percentage_contribution := intrinsic_percentages[i]:
+                percentage_contributions.append(percentage_contribution)
+                continue
+            for row, row_origins in zip(grid, rows_origins):
+                if (origin := row_origins[i]) is None:
                     continue
-                origin = max(indexes)
-                origin_cell = grid[j][origin]
+                origin_cell = row[origin]
                 if origin_cell.colspan - 1 != span:
                     continue
-                cell_slice = slice(origin, origin + origin_cell.colspan)
-                baseline_percentage = sum(intrinsic_percentages[cell_slice])
-
-                # Cell contribution to intrinsic percentage width
-                if intrinsic_percentages[i] == 0:
-                    diff = max(
-                        0,
-                        _percentage_contribution(origin_cell) -
-                        baseline_percentage)
-                    other_columns_contributions = [
-                        max_content_widths[j]
-                        for j in range(
-                            origin, origin + origin_cell.colspan)
-                        if intrinsic_percentages[j] == 0]
-                    other_columns_contributions_sum = sum(
-                        other_columns_contributions)
-                    if other_columns_contributions_sum == 0:
-                        if other_columns_contributions:
-                            ratio = 1 / len(other_columns_contributions)
-                        else:
-                            ratio = 1
-                    else:
-                        ratio = (
-                            max_content_widths[i] /
-                            other_columns_contributions_sum)
-                    percentage_contribution = max(
-                        percentage_contribution,
-                        diff * ratio)
+                cell_percentage_contribution = get_percentage_contribution(
+                    origin_cell, origin, max_content_widths[i])
+                percentage_contribution = max(
+                    percentage_contribution, cell_percentage_contribution)
 
             percentage_contributions.append(percentage_contribution)
 
@@ -556,6 +575,7 @@ def table_and_columns_preferred_widths(context, box, outer=True):
         for cell in zipped_grid[i]:
             if (cell and cell.colspan == 1 and
                     cell.style['width'] != 'auto' and
+                    not check_math(cell.style['width']) and
                     cell.style['width'].unit != '%'):
                 constrainedness[i] = True
                 break
@@ -622,11 +642,11 @@ def table_and_columns_preferred_widths(context, box, outer=True):
     table_min_content_width = (
         total_horizontal_border_spacing + sum(min_content_widths))
     table_max_content_width = (
-        total_horizontal_border_spacing + max(
-            [sum(max_content_widths), large_percentage_contribution] +
-            small_percentage_contributions))
+        total_horizontal_border_spacing + max([
+            sum(max_content_widths), large_percentage_contribution,
+            *small_percentage_contributions]))
 
-    if table.style['width'] != 'auto' and table.style['width'].unit == 'px':
+    if table.style['width'] != 'auto' and table.style['width'].unit.lower() == 'px':
         # "percentages on the following properties are treated instead as
         # though they were the following: width: auto"
         # https://dbaron.org/css/intrinsic/#outer-intrinsic
@@ -650,10 +670,8 @@ def table_and_columns_preferred_widths(context, box, outer=True):
         min_content_widths, max_content_widths, intrinsic_percentages,
         constrainedness, total_horizontal_border_spacing, zipped_grid)
     context.tables[table] = result = {
-        False: (table_min_content_width, table_max_content_width) + result,
-        True: (
-            (table_outer_min_content_width, table_outer_max_content_width) +
-            result),
+        False: (table_min_content_width, table_max_content_width, *result),
+        True: (table_outer_min_content_width, table_outer_max_content_width, *result),
     }
     return result[outer]
 
@@ -666,10 +684,9 @@ def replaced_min_content_width(box, outer=True):
         if height == 'auto' or height.unit == '%':
             height = 'auto'
         else:
-            assert height.unit == 'px'
+            assert height.unit.lower() == 'px'
             height = height.value
-        if (box.style['max_width'] != 'auto' and
-                box.style['max_width'].unit == '%'):
+        if box.style['max_width'] != 'auto' and box.style['max_width'].unit == '%':
             # See https://drafts.csswg.org/css-sizing/#intrinsic-contribution
             width = 0
         else:
@@ -679,12 +696,12 @@ def replaced_min_content_width(box, outer=True):
                     box.style['image_resolution'], box.style['font_size']))
             width, _ = default_image_sizing(
                 intrinsic_width, intrinsic_height, intrinsic_ratio, 'auto',
-                height, default_width=300, default_height=150)
+                height, default_width=0, default_height=0)
     elif box.style['width'].unit == '%':
         # See https://drafts.csswg.org/css-sizing/#intrinsic-contribution
         width = 0
     else:
-        assert width.unit == 'px'
+        assert width.unit.lower() == 'px'
         width = width.value
     return adjust(box, outer, width)
 
@@ -697,7 +714,7 @@ def replaced_max_content_width(box, outer=True):
         if height == 'auto' or height.unit == '%':
             height = 'auto'
         else:
-            assert height.unit == 'px'
+            assert height.unit.lower() == 'px'
             height = height.value
         image = box.replacement
         intrinsic_width, intrinsic_height, intrinsic_ratio = (
@@ -710,7 +727,7 @@ def replaced_max_content_width(box, outer=True):
         # See https://drafts.csswg.org/css-sizing/#intrinsic-contribution
         width = 0
     else:
-        assert width.unit == 'px'
+        assert width.unit.lower() == 'px'
         width = width.value
     return adjust(box, outer, width)
 
@@ -750,17 +767,30 @@ def trailing_whitespace_size(context, box):
     """Return the size of the trailing whitespace of ``box``."""
     from .inline import split_first_line, split_text_box
 
+    # Find last box child, keep last parent to remove nested trailing spaces.
+    last_parent = None
     while isinstance(box, (boxes.InlineBox, boxes.LineBox)):
         if not box.children:
             return 0
-        box = box.children[-1]
-    if not (isinstance(box, boxes.TextBox) and box.text and
-            box.style['white_space'] in ('normal', 'nowrap', 'pre-line')):
+        last_parent, box = box, box.children[-1]
+
+    # Return early if possible.
+    if not isinstance(box, boxes.TextBox) or not box.text:
+        # There’s no text in last child.
         return 0
-    stripped_text = box.text.rstrip(' ')
-    if box.style['font_size'] == 0 or len(stripped_text) == len(box.text):
+    elif box.style['white_space'] not in ('normal', 'nowrap', 'pre-line'):
+        # Spaces don’t collapse.
         return 0
-    if stripped_text:
+    elif box.style['font_size'] == 0:
+        # Trailing spaces take no space.
+        return 0
+    elif not box.text.endswith(' '):
+        # No trailing space.
+        return 0
+
+    # Strip text.
+    if stripped_text := box.text.rstrip(' '):
+        # Stripped text is not empty, calculate width difference.
         resume = 0
         while resume is not None:
             old_resume = resume
@@ -770,12 +800,17 @@ def trailing_whitespace_size(context, box):
         stripped_box, resume, _ = split_text_box(
             context, stripped_box, None, old_resume)
         if stripped_box is None:
-            # old_box split just before the trailing spaces
+            # Old box is split just before the trailing spaces.
             return old_box.width
         else:
+            # Return difference between old width and stripped width.
             assert resume is None
             return old_box.width - stripped_box.width
     else:
+        # Stripped text is empty, render spaces to get width.
         _, _, _, width, _, _ = split_first_line(
             box.text, box.style, context, None, box.justification_spacing)
+        # Remove possible trailing spaces from previous child.
+        if last_parent and len(last_parent.children) >= 2:
+            width += trailing_whitespace_size(context, last_parent.children[-2])
         return width
